class Tfibpair
{
 private:
  mpz_t x1,x2;
  mpz_t h1,h2,h3;
 public:
  Tfibpair(const unsigned int f1=0, const unsigned int f0=1)
   {
     mpz_init_set_ui(x1,f1);
     mpz_init_set_ui(x2,f0);
     mpz_init(h1); mpz_init(h2); mpz_init(h3);
   }
  Tfibpair(const Tfibpair &X)
   {
     mpz_init_set(x1,X.x1);
     mpz_init_set(x2,X.x2);
     mpz_init(h1); mpz_init(h2); mpz_init(h3);
   }
  ~Tfibpair()
   {
     mpz_clear(h3); mpz_clear(h2); mpz_clear(h1);
     mpz_clear(x2); mpz_clear(x1);
   }

  inline void set(const unsigned int f1, const unsigned int f0)
   {
     mpz_set_ui(x1,f1);
     mpz_set_ui(x2,f0);
   }

  inline void set(const Tfibpair &X)
   {
     mpz_set(x1,X.x1);
     mpz_set(x2,X.x2);
     // Hilfsvariablen bleiben unbercksichtigt!
   }

  inline const mpz_t& fn(void) const { return x1; }
  inline const mpz_t& fnm1(void) const { return x2; }
  inline const mpz_t& Ln(void) { mpz_add(h1,x2,x2); mpz_add(h1,h1,x1); return h1; }

  inline void mod(const mpz_t &m)
   {
     mpz_mod(x1,x1,m);
     mpz_mod(x2,x2,m);
   }
  inline void mul(const Tfibpair &Q)
   {
     mpz_add(h1,x1,x2); mpz_add(h2,Q.x1,Q.x2); mpz_mul(h1,h1,h2);
     mpz_mul(h2,x1,Q.x1); 
     mpz_mul(h3,x2,Q.x2); 
     mpz_sub(x1,h1,h3); mpz_add(x2,h2,h3);
   }
  inline void square() 
   {
     mpz_add(h1,x1,x2); mpz_mul(h1,h1,h1);
     mpz_mul(h2,x1,x1); 
     mpz_mul(h3,x2,x2); 
     mpz_sub(x1,h1,h3); mpz_add(x2,h2,h3);
   }
  inline void fastsquare()
   {
     // assumes that first index is already even
#ifdef DEBUG
     const bool flag = (mpz_cmp_ui(x1,1)==0 && mpz_cmp_ui(x2,0)==0);
     if (flag) cout << "error, error!" << endl;
#endif
     mpz_mul(h1,x1,x1); mpz_mul(h2,x2,x2);
     mpz_add(x2,h1,h2);
     mpz_mul_2exp(h1,h1,2); mpz_sub(h1,h1,h2); mpz_add_ui(h1,h1,2);
     mpz_sub(x1,h1,x2);
   }
  inline void pow(const Tfibpair &FQ, const unsigned int n)
   {
     Tfibpair Q(FQ);

     if (n&1)
      { if (&FQ!=this) set(FQ); }	
     else
      { mpz_set_ui(x1,0); mpz_set_ui(x2,1); }

     for (unsigned int i=2; i<=n; i<<=1)
      {
        Q.square();
        if (n&i) mul(Q);
      }
   }
  inline void powmod(const Tfibpair &FQ, const unsigned int n, const mpz_t &m)
   {
     Tfibpair Q(FQ);

     if (n&1)
      { if (&FQ!=this) set(FQ); }	
     else
      { mpz_set_ui(x1,0); mpz_set_ui(x2,1); }

     for (unsigned int i=2; i<=n; i<<=1)
      {
        Q.square(); Q.mod(m);
        if (n&i) { mul(Q); mod(m); }
      }
   }
  inline void fastpowmod(const Tfibpair &FQ, const unsigned int n, const mpz_t &m)
   {
     Tfibpair Q(FQ);

     if (n&1)
      { if (&FQ!=this) set(FQ); }	
     else
      { mpz_set_ui(x1,0); mpz_set_ui(x2,1); }

     for (unsigned int i=2; i<=n; i<<=1)
      {
        Q.fastsquare(); Q.mod(m);
        if (n&i) { mul(Q); mod(m); }
      }
   }
  inline void step_forward(unsigned int d)
   {
     // (F[n],F[n-1]) -> (F[n+d],F[n+d-1])
     while(d--) { mpz_swap(x1,x2); mpz_add(x1,x1,x2); }
     // one can also use F[n+d] = F[d+1]*F[n] + F[d]*F[n-1]
     // not yet implemented...
   }
  inline void step_backward(unsigned int d)
   {
     // (F[n],F[n-1]) -> (F[n-d],F[n-d-1])
     while(d--) { mpz_sub(x1,x1,x2); mpz_swap(x1,x2); }
     // one can also use F[n+d] = F[d+1]*F[n] + F[d]*F[n-1]
     // not yet implemented...
   }
};


void fibonacci_ppm1(const int phase1, const double phase2)
{
  // Fibonacci Methode modifiziert in drei Phasen:
  // Phase 0: alle kleinen Zahlen bis konstanter Obergrenze
  // Phase 1: (im wesentlichen) Primzahlen bis phase1 als Obergrenze
  //          (Primfaktor+-1) kann/darf in alle diese Primzahlen zerfallen
  // Phase 2: (im wesentlichen) Primzahlen bis phase2 als Obergrenze
  //          bis auf einen(!) aus Phase 2 mssen alle anderen
  //          Faktoren von (Primfaktor-1) bereits vorher entdeckt worden
  //          sein, um Primfaktor entdecken zu knnen.
  //          Es stehen (theoretisch) diverse Methoden fr Phase 2 zur Verfgung,
  //          jedoch habe ich mir nur die effizienteste aus der (p-1)-Methode
  //          fr "fibonacci" angepat:
  //           - improved standard continuation with pairing & fft

  Tfibpair F(1,0);
  vector<int> KnownIndices;
  int VorigerIndex = 0;

  int i;

  // Phase 0: kleine Zahlen (mit Primzahlpotenzen)
  cout << "Fibonacci-Phase 0" << endl;
  F.square();

  for (i=3; i<10000; i++)
    {
      F.fastpowmod(F,i,n);
retry:
      mpz_gcd(x,F.fn(),n);
      if (mpz_cmp_ui(x,1)!=0)  /* Teiler gefunden? */
	{
	  cout << "Faktor gefunden in Phase 0: i=" << i << endl;
inner_retry:
	  if (mpz_cmp(x,n)==0) // gefundener Faktor = n ?
	    {
	      cout << "Gefundene Faktorisierung ist trivial." << endl;

              int h=i; int j=2;
	      while (h>j) // grten Primteiler ermitteln 
               {
                 while(h>j && h%j==0) h/=j;
		 ++j;
               }

              if (KnownIndices.empty()) KnownIndices.push_back(h);
#if 0
	      else if (i>VorigerIndex) // im Normalfall (auer fr Teiler von Lucaszahlen) sollten die Indices nicht mehr ansteigen...
                { KnownIndices.back()=VorigerIndex/KnownIndices.back(); }
#endif
	      else // war dieser bereits im vorigen Durchlauf hinzugefgt
               if (VorigerIndex!=i || KnownIndices.back()!=h) KnownIndices.push_back(h);
               else
                {
                  // dann ggf. wieder entfernen
		  if (KnownIndices.back()!=i/h) KnownIndices.pop_back();
		  // und durch anderen Teiler ersetzen
		  KnownIndices.push_back(i/h);
                }
	      VorigerIndex=i;

	      cout << "KnownIndices:";
              for (vector<int>::iterator p=KnownIndices.begin(); p!=KnownIndices.end(); ++p) cout << " " << *p;
              cout << endl;

              if (KnownIndices.empty() || KnownIndices.back()==1)
               {
#if 1
		 F.set(1,0);
		 for (vector<int>::iterator p=KnownIndices.begin(); p!=KnownIndices.end(); ++p) F.powmod(F,*p,n);

	         mpz_gcd(x,F.fn(),n);
	         if (mpz_cmp_ui(x,1)!=0)
                  {
		    if (mpz_cmp(x,n)!=0)
		     {
                       cout << "Fibonacciwert sieht gut aus!" << endl;
                       F.set(1,0); F.square(); i=2;
		       goto inner_retry;
		     }
		    else cout << "Fibonacciwert ist trivial..." << endl;
		  }
#endif
                 cout << "Gebe auf fr " << n << endl;
                 // { char ch; cin >> ch; }

                 goto fertig; // trivialer Teiler -> Faktoren anders aufrollen oder Verfahren abbrechen
               }

              if (KnownIndices.size()<100)
               {
		 F.set(1,0); F.square();
		 for (vector<int>::iterator p=KnownIndices.begin(); p!=KnownIndices.end(); ++p) F.fastpowmod(F,*p,n);
		 cout << "Versuche es erneut..." << endl;
		 i=2; goto retry;
               }
              cout << "Breche ab fr " << n << endl;
              // { char ch; cin >> ch; } 

	      goto fertig; // trivialer Teiler -> Faktoren anders aufrollen oder Verfahren abbrechen
	    }
	  if (mpz_probab_prime_p(x,probab_prime_checks))
	    {
	      cout << x << " ist Faktor." << endl;
	      Factorization_to_file << MAL() << x << " [fib0]" << flush;
	    }
	  else
	    {
	      cout << x << " ist zusammengesetzter Faktor." << endl;
              mpz_divexact(n,n,x);
              if (mpz_probab_prime_p(n,probab_prime_checks))
               {
                 mpz_swap(n,x);
                 cout << x << " ist Faktor." << endl;
	         Factorization_to_file << MAL() << x << " [fib0/factorswap]" << flush;
		 //goto fertig;
               }
              else
	       {
		mpz_t saved_n;
		mpz_init_set(saved_n,n);
		mpz_swap(n,x);
		cout << "Rufe fibonacci rekursiv auf..." << endl;
                fibonacci_ppm1(phase1,phase2);
		cout << "zurck aus Rekursion. Setze Fibonacci fort..." << endl;
		if (mpz_probab_prime_p(n,probab_prime_checks))
                 Factorization_to_file << MAL() << n << " [fib0]" << flush;
		else
                 Factorization_to_file << MAL() << n << " [fib0] [composite]" << flush;
		mpz_swap(n,saved_n); mpz_clear(saved_n);
	       }
              mpz_set_ui(x,1);
              F.set(1,0); F.square(); i=2;
	    }
	  KnownIndices.clear(); VorigerIndex=0;
	  mpz_divexact(n,n,x);
	  if (mpz_probab_prime_p(n,probab_prime_checks)) goto fertig;
	}
    }
  
#if 1 /* normale Fibonacci-Phase 1 */
  {
   // Phase 1: Primzahlen ohne Potenzen (aber knnen in beliebiger Kombination p-1 teilen)
   const int D=2*3*5*7*11*13;
   i=i-(i%D); // i mit Eigenschaft i=5 (mod 6) fr schnellere Primzahlermittlung
   cout << "Fibonacci-Phase 1" << endl;
   while (i<=phase1)
    {
      // Primzahlen aussieben
      bool sieve[D];
      for (int j=1; j<=D; j+=2) sieve[j]=true; // Sieb initialisieren
      for (int p=5, dp=2; p*p<D+i; p+=dp, dp=6-dp)
       for (int j=p-(i%p); j<D; j+=p) sieve[j]=false;
       /* 
         Bemerkung:
          a) p*p<D+i ist als Primzahlbedingung fr das Intervall
             etwas zu streng, aber strt wegen Phase 0 nicht.
             (-> Primzahlen unter sqrt(D) werden nicht erkannt) 
          b) p-(i%p) ist zwar p, wenn p teilt i;
             aber: den Index 0 brauchen wir spter nicht!
       */
            
      for (int j=1, dj=4; j<D; j+=dj, dj=6-dj)
       {
         //if (sieve[j]!=is_prime(i+j))
         //  { cerr << "keine bereinstimmung: " << i+j << endl; }
         if (sieve[j]) F.fastpowmod(F,i+j,n);
       }
      // Schleife vermeidet hufige gcd-Berechnung
      // in seltenen Fllen knnen aber mehrere Faktoren zusammenfallen
      // -> fibonacci-Methode versagt dann (ggf. Schleifendurchlufe vermindern)

      i+=D;
      cout << "Phase1: " << i << "/" << phase1 << "\r" << flush;

      mpz_gcd(x,F.fn(),n);
      if (mpz_cmp_ui(x,1)!=0)
        {	  
	  cout << "Faktor gefunden in Phase 1: i=" << i << endl;
	  if (mpz_cmp(x,n)==0) // gefundener Faktor = n ?
	    {
	      cout << "Gefundene Faktorisierung ist trivial. Breche ab." << endl;
	      goto fertig; // trivialer Teiler -> Faktoren anders aufrollen oder Verfahren abbrechen
            }
	  if (mpz_probab_prime_p(x,probab_prime_checks))
	    {
	      cout << x << " ist Faktor." << endl;
	      Factorization_to_file << MAL() << x << " [fib1]" << flush;
	    }
	  else
	    {
	      cout << x << " ist zusammengesetzter Faktor." << endl;
              mpz_divexact(n,n,x);
              if (mpz_probab_prime_p(n,probab_prime_checks))
               {
                 mpz_swap(n,x);
	         Factorization_to_file << MAL() << x << " [fib1/factorswap]" << flush;
		 goto fertig;
               }
              else
                Factorization_to_file << MAL() << x << " [fib1] [composite]" << flush;
              mpz_set_ui(x,1);
 	    }
	  mpz_divexact(n,n,x);
	  if (mpz_probab_prime_p(n,probab_prime_checks)) goto fertig;
	}
    }
   i--; int d=4; while (!is_prime(i)) { i-=d; d=6-d; } // i auf vorige Primzahl adjustieren
   cout << "Fibonacci-Phase 1 bis " << i << endl;
   if (i>=phase2) goto fertig;
  }
#endif

#if 1 /* improved standard continuation with pairing & fft */
  {
   double ii=i;
   //const double i = -1; // uncomment this just to be sure that i isn't used anymore...

   // Phase 2: isolierte (Prim-)zahl: p-1 mu bis auf diese Zahl vorher vollstndig zerlegbar gewesen sein.
   cout << "Fibonacci-Phase 2 (improved with pairing & fft)" << endl;

   /* zur Bedeutung nachfolgender Werte siehe elliptische Kurven,
      fft-continuation */
   unsigned int pg=256, pms=2*1050;
   get_fft_parameter(phase2,pg,pms); // Parameter (pg,pms) fr "phase2" holen

   const int D = pms;
   const int Polynom_Index = pg;

   cout << "Polynomgre=" << Polynom_Index << endl;
   cout << "Einzel-Intervallsize=" << D << endl;

   mpz_t sammel[Polynom_Index];
   for (int j=0; j<Polynom_Index; ++j) mpz_init(sammel[j]);

  /* Grundidee:
     Ausnutzung der folgenden Gleichheit:
     F[i+d] * F[i-d] = F[i]^2 + F[d]^2  (fr i gerade, d ungerade) 
     
     In Intervallen der Gre D wird nun i erhht.
       -> Vorberechnung von F[d]^2 fr 0<d<D,
       -> auerdem: pro Intervall eine Bestimmung von F[i]^2

    Bemerkung:
      Die hierauf aufbauende Paarbildungsmethode konstruiert nicht
      die durch andere Methoden erreichbaren Primzahlpaare, sie ist
      aber dafr vergleichsweise einfach zu implementieren und sie
      kommt mit weniger Speicherplatz als die improved-standard-continuation
      aus. Sobald mehrere Primzahlpaare innerhalb eines Intervalls
      auftauchen, ist sie schneller als die improved standard continuation.
  */

   // Precomputing most common used values:
   int Polynom_Index_i = 0;

   F.square(); F.mod(n); // um wirklich sicherzustellen, da Index gerade ist...

   Tfibpair stepF(F), mulF(F);
   mulF.square(); mulF.mod(n); 
   for (int j=1; j<D/2; j+=2)
    {
      if (gcd(j,D)==1)
	{
          mpz_set(x,stepF.fn()); mpz_mul(x,x,x); mpz_mod(x,x,n);
          mpz_set(sammel[Polynom_Index_i++],x);
        }
      stepF.mul(mulF); stepF.mod(n);
    }

   cout << "unvergrerte Polynomgre: " << Polynom_Index_i << endl;
   polynomial::TPolynom Polynom = NULL; // hier steht das anschlieend konstruierte Polynom drin
   // Rest bis zur nchsten Zweierpotenz mit Nullen auffllen
   /* Bemerkung: Besser wre es, Punkte auerhalb des Suchbereichs
                 zu whlen (Motto: doppeltes, aber auerhalb des
                 Suchbereichs lchriges Netz)
    */
   { 
     int anz_Punkte=Polynom_Index_i;    
     while (anz_Punkte<Polynom_Index) mpz_set_ui(sammel[anz_Punkte++],0);
     Polynom_Index_i=polynomial::Polynom_aus_Nullstellen_konstruieren(Polynom,sammel,anz_Punkte,n);
     cout << "Polynom erzeugt. Polynomindex = " << Polynom_Index_i << endl;
   }

   ii=floor(ii/D)*D; // i=i-(i%D); // Startpunkt ermitteln

   mpz_t Fii, Lii, Liip, FD, LD;
   {
     Tfibpair h(F); // wird nur temporr bentigt...
     h.powmod(h,D,n);
     mpz_init_set(FD,h.fn()); // Lucaswert L[D]
     mpz_init_set(LD,h.Ln()); // Fibonacciwert F[D]

     Tfibpair hp(h);
     int pow = static_cast<int>(ii/D);
     h.powmod(h,pow,n);
     mpz_init_set(Fii,h.fn());
     mpz_init_set(Lii,h.Ln());

     if (--pow<0) pow=-pow; 
     hp.powmod(hp,pow,n);
     mpz_init_set(Liip,hp.Ln()); // Lucaswert L[ii-D]
   }

   ii-=D/2; // Intervall von [i-D/2,i+D/2] auf [i,i+D] normieren
   while (ii<=phase2)
    {
      cout << "Phase2: " << ii << "/" << phase2 << endl << flush;
      cout << "Sammle Werte..." << endl;

      //Werte sammeln
      int sammel_index=0;
      while (sammel_index<(Polynom_Index_i-1)/2)
       {
	 mpz_mul(x,Fii,Fii); mpz_mod(x,x,n);
         mpz_set(sammel[sammel_index++],x);
         // Werte fr nchstes Intervall bestimmen...
         ii+=D; // neues Intervall (bis i haben wir nun abgearbeitet)
	 
         // F[ii+D]=(F[ii]*L[D]+F[D]*L[ii])/2
         // L[ii+D]=L[ii]*L[D]-L[ii-D]

	 mpz_mul(x,Fii,LD); mpz_mul(y,Lii,FD);
         mpz_add(Fii,x,y); mpz_mod(Fii,Fii,n);
	 if (mpz_odd_p(Fii)) mpz_add(Fii,Fii,n);
         mpz_tdiv_q_2exp(Fii,Fii,1);

	 mpz_mul(x,Lii,LD); mpz_sub(x,x,Liip); mpz_mod(Liip,x,n);
	 mpz_swap(Liip,Lii);
       }

      // Polynomauswertung starten
      cout << "Phase2: " << ii << "/" << phase2 << endl << flush;
      cout << "starte Polynomauswertung" << endl;
      cout << "Parameter: Polynomindex: " << Polynom_Index_i
           << ", Anzahl der Punkte: " << sammel_index << endl;
      polynomial::multipoint_eval(sammel,Polynom,Polynom_Index_i,sammel,sammel_index,n);
      cout << "Polynomauswertung beendet. Teste auf ggT." << endl;
      mpz_set(y,sammel[0]);
      for (int j=1; j<sammel_index; ++j)
        { mpz_mul(y,y,sammel[j]); mpz_mod(y,y,n); }
      // nun (erst) den ggT ermitteln
      mpz_gcd(x,y,n);
      if (mpz_cmp_ui(x,1)==0) continue;
      
      // nun ist/sind ein/mehrere Teiler(in/nen) gefunden,
      // mgliches Problem: vielleicht mehrere Teiler akkumuliert.
      // Da es aber nur log-viele Teiler geben kann, ist dieser
      // Abschnitt nicht zeitkritisch; also strt es wenig, alle
      // Werte nochmal durchzugehen (dies senkt die Wahrscheinlichkeit,
      // akkumulierte Teiler zu erhalten).
    
      for (int j=1; j<sammel_index; ++j)
       {
	 mpz_gcd(x,sammel[j],n);
         if (mpz_cmp_ui(x,1)!=0)
	  {
	   cout << "Faktor gefunden in Phase 2: i=" << ii << endl;
	   if (mpz_probab_prime_p(x,probab_prime_checks))
	    {
	      cout << x << " ist Faktor." << endl;
	      Factorization_to_file << MAL() << x << " [fib2]" << flush;
	    }
	   else
	    {
	      cout << x << " ist zusammengesetzter Faktor." << endl;
	      Factorization_to_file << MAL() << x << " [fib2] [composite]" << flush;
	    }
	   mpz_divexact(n,n,x);
	   if (mpz_probab_prime_p(n,probab_prime_checks)) goto fertig_phase2;
           // anschlieend noch Polynom vereinfachen:
	   for (int j=0; j<Polynom_Index_i; ++j) mpz_mod(Polynom[j],Polynom[j],n);
	  }
       }

    }

 fertig_phase2:
  cout << "Fibonacci-Phase 2 bis " << ii << endl;
  mpz_clear(Fii); mpz_clear(Lii); mpz_clear(Liip);
  mpz_clear(FD); mpz_clear(LD);
  for (int j=0; j<Polynom_Index; ++j) mpz_clear(sammel[j]);
  for (int j=0; j<Polynom_Index_i; ++j) mpz_clear(Polynom[j]);
  delete [] Polynom;
  }
#endif


 fertig:
  cout << "Fibonacci abgeschlossen." << endl;
}
